from terminaltables import AsciiTable
from datetime import datetime
from pypeerassets import Deck, CardTransfer
from pypeerassets.pautils import exponent_to_amount


def tstamp_to_iso(tstamp):
    '''make iso timestamp from unix timestamp'''

    return datetime.fromtimestamp(tstamp).isoformat()


def print_table(title, heading, data):
    " prints a table to the terminal using terminaltables.AsciiTable "

    data = list(data)
    data.insert(0, heading)
    table = AsciiTable(data, title=title)
    print(table.table)


def deck_title(deck):
    return "Deck ID: " + deck.id + " "


def deck_summary_line_item(deck: Deck):

    d = deck.__dict__
    return [d["id"],
            d["name"],
            d["issuer"],
            d["issue_mode"],
            d["tx_confirmations"]
            ]


def print_deck_list(decks: list):
    '''Show summary of every deck'''

    print_table(
            title="Decks",
            heading=("ID", "name", "issuer", "mode", "confirms"),
            data=map(deck_summary_line_item, decks))


def print_deck_info(deck: Deck):

    deck.issue_time = tstamp_to_iso(deck.issue_time)
    deck.data = str(deck.asset_specific_data)

    print_table(
            title=deck_title(deck),
            heading=("name", "issuer", "issue mode", "decimals", "confirms", "timestamp", "data"),
            data=[[
                getattr(deck, attr) for attr in
                        ["name", "issuer", "issue_mode",
                         "number_of_decimals", "tx_confirmations",
                         "issue_time", "data"]]]
                         )


def card_line_item(card: CardTransfer):

    c = card.__dict__
    return [c["txid"],
            c["tx_confirmations"],
            c['cardseq'],
            c["sender"],
            c["receiver"][0],
            exponent_to_amount(c["amount"][0], c["number_of_decimals"]),
            c["type"]
            ]


def print_card_list(cards: list):

    print_table(
            title="Card transfers of deck {deck}:".format(deck=cards[0].deck_id),
            heading=("txid", "confirms", "seq", "sender", "receiver", "amount", "type"),
            data=map(card_line_item, cards))



from typing import Optional
import fire
import random
import pypeerassets as pa
import json

from pypeerassets.pautils import (amount_to_exponent,
                                  exponent_to_amount,
                                  parse_card_transfer_metainfo,
                                  parse_deckspawn_metainfo
                                  )
from pypeerassets.transactions import NulldataScript

from pacli.provider import provider
from pacli.config import Settings
from pacli.keystore import init_keystore
from pacli.tui import print_deck_info, print_deck_list
from pacli.tui import print_card_list
from pacli.export import export_to_csv
from pacli.utils import (cointoolkit_verify,
                         signtx,
                         sendtx
                         )
from pacli.coin import Coin


class Address:

    '''my personal address'''

    def show(self, pubkey: bool=False, privkey: bool=False, wif: bool=False) -> str:
        '''print address, pubkey or privkey'''

        if pubkey:
            return Settings.key.pubkey
        if privkey:
            return Settings.key.privkey
        if wif:
            return Settings.key.wif

        return Settings.key.address

    @classmethod
    def balance(self) -> float:

        return float(provider.getbalance(Settings.key.address))

    def derive(self, key: str) -> str:
        '''derive a new address from <key>'''

        return pa.Kutil(Settings.network, from_string=key).address

    def random(self, n: int=1) -> list:
        '''generate <n> of random addresses, useful when testing'''

        return [pa.Kutil(network=Settings.network).address for i in range(n)]

    def get_unspent(self, amount: int) -> Optional[dict]:
        '''quick find UTXO for this address'''

        try:
            return provider.select_inputs(Settings.key.address, 0.02)['utxos'][0].__dict__['txid']
        except KeyError:
            print({'error': 'No UTXOs ;('})


class Deck:

    @classmethod
    def list(self):
        '''find all valid decks and list them.'''

        decks = pa.find_all_valid_decks(provider, Settings.deck_version,
                                        Settings.production)

        print_deck_list(decks)

    @classmethod
    def find(self, key):
        '''
        Find specific deck by key, with key being:
        <id>, <name>, <issuer>, <issue_mode>, <number_of_decimals>
        '''

        decks = pa.find_all_valid_decks(provider,
                                        Settings.deck_version,
                                        Settings.production)
        print_deck_list(
            (d for d in decks if key in d.id or (key in d.to_json().values()))
            )

    @classmethod
    def info(self, deck_id):
        '''display deck info'''

        deck = pa.find_deck(provider, deck_id, Settings.deck_version,
                            Settings.production)
        print_deck_info(deck)

    @classmethod
    def p2th(self, deck_id: str) -> str:
        '''print out deck p2th'''

        return pa.Kutil(network=Settings.network,
                        privkey=bytearray.fromhex(deck_id)).address

    @classmethod
    def __new(self, name: str, number_of_decimals: int, issue_mode: int,
              asset_specific_data: str=None, locktime=None):
        '''create a new deck.'''

        network = Settings.network
        production = Settings.production
        version = Settings.deck_version

        new_deck = pa.Deck(name, number_of_decimals, issue_mode, network,
                           production, version, asset_specific_data)

        return new_deck

    @classmethod
    def spawn(self, verify: bool=False, sign: bool=False,
              send: bool=False, locktime: int=0, **kwargs):
        '''prepare deck spawn transaction'''

        deck = self.__new(**kwargs)

        spawn = pa.deck_spawn(provider=provider,
                              inputs=provider.select_inputs(Settings.key.address, 0.02),
                              deck=deck,
                              change_address=Settings.change,
                              locktime=locktime
                              )

        if verify:
            return cointoolkit_verify(spawn.hexlify())  # link to cointoolkit - verify

        if sign:

            tx = signtx(spawn)

            if send:
                return sendtx(tx)

            return tx.hexlify()

        return spawn.hexlify()

    @classmethod
    def encode(self, json: bool=False, **kwargs) -> str:
        '''compose a new deck and print out the protobuf which
           is to be manually inserted in the OP_RETURN of the transaction.'''

        if json:
            return self.__new(**kwargs).metainfo_to_dict

        return self.__new(**kwargs).metainfo_to_protobuf.hex()

    @classmethod
    def decode(self, hex: str) -> dict:
        '''decode deck protobuf'''

        script = NulldataScript.unhexlify(hex).decompile().split(' ')[1]

        return parse_deckspawn_metainfo(bytes.fromhex(script),
                                        Settings.deck_version)

    def issue_modes(self):

        im = tuple({mode.name: mode.value} for mode_name, mode in pa.IssueMode.__members__.items())

        print(json.dumps(im, indent=1, sort_keys=True))

    def my(self):
        '''list decks spawned from address I control'''

        return self.find(Settings.key.address)


class Card:

    '''card information and manipulation'''

    @classmethod
    def __find_deck(self, deckid) -> Deck:

        deck = pa.find_deck(provider, deckid,
                            Settings.deck_version,
                            Settings.production)

        if deck:
            return deck

    @classmethod
    def __list(self, deckid: str):

        deck = self.__find_deck(deckid)

        try:
            cards = pa.find_all_valid_cards(provider, deck)
        except pa.exceptions.EmptyP2THDirectory as err:
            return err

        return {'cards': list(cards),
                'deck': deck}

    @classmethod
    def list(self, deckid: str):
        '''list the valid cards on this deck'''

        cards = self.__list(deckid)['cards']

        print_card_list(cards)

    def balances(self, deckid: str):
        '''list card balances on this deck'''

        cards, deck = self.__list(deckid).values()

        state = pa.protocol.DeckState(cards)

        balances = [exponent_to_amount(i, deck.number_of_decimals)
                    for i in state.balances.values()]

        print(json.dumps(dict(zip(state.balances.keys(), balances)
                              ), indent=4))

    def checksum(self, deckid: str) -> bool:
        '''show deck card checksum'''

        cards, deck = self.__list(deckid).values()

        state = pa.protocol.DeckState(cards)

        return state.checksum

    @staticmethod
    def to_exponent(number_of_decimals, amount):
        '''convert float to exponent'''

        return amount_to_exponent(amount, number_of_decimals)

    @classmethod
    def __new(self, deckid: str, receiver: list=None,
              amount: list=None, asset_specific_data: str=None) -> pa.CardTransfer:
        '''fabricate a new card transaction
        * deck_id - deck in question
        * receiver - list of receivers
        * amount - list of amounts to be sent, must be float
        '''

        deck = self.__find_deck(deckid)

        if isinstance(deck, pa.Deck):
            card = pa.CardTransfer(deck=deck,
                                   receiver=receiver,
                                   amount=[self.to_exponent(deck.number_of_decimals, i)
                                           for i in amount],
                                   version=deck.version,
                                   asset_specific_data=asset_specific_data
                                   )

            return card

        raise Exception({"error": "Deck {deckid} not found.".format(deckid=deckid)})

    @classmethod
    def transfer(self, deckid: str, receiver: list=None, amount: list=None,
                 asset_specific_data: str=None,
                 locktime: int=0, verify: bool=False,
                 sign: bool=False, send: bool=False) -> str:
        '''prepare CardTransfer transaction'''

        card = self.__new(deckid, receiver, amount, asset_specific_data)

        issue = pa.card_transfer(provider=provider,
                                 inputs=provider.select_inputs(Settings.key.address, 0.02),
                                 card=card,
                                 change_address=Settings.change,
                                 locktime=locktime
                                 )

        if verify:
            return cointoolkit_verify(issue.hexlify())  # link to cointoolkit - verify

        if sign:

            tx = signtx(issue)

            if send:
                return sendtx(tx)

            return tx.hexlify()

        return issue.hexlify()

    @classmethod
    def burn(self, deckid: str, receiver: list=None, amount: list=None,
             asset_specific_data: str=None,
             locktime: int=0, verify: bool=False, sign: bool=False) -> str:
        '''wrapper around self.transfer'''

        return self.transfer(deckid, receiver, amount, asset_specific_data,
                             locktime, verify, sign)

    @classmethod
    def issue(self, deckid: str, receiver: list=None, amount: list=None,
              asset_specific_data: str=None,
              locktime: int=0, verify: bool=False, sign: bool=False) -> str:
        '''Wrapper around self.tranfer'''

        return self.transfer(deckid, receiver, amount, asset_specific_data,
                             locktime, verify, sign)

    @classmethod
    def encode(self, deckid: str, receiver: list=None, amount: list=None,
               asset_specific_data: str=None, json: bool=False) -> str:
        '''compose a new card and print out the protobuf which
           is to be manually inserted in the OP_RETURN of the transaction.'''

        card = self.__new(deckid, receiver, amount, asset_specific_data)

        if json:
            return card.metainfo_to_dict

        return card.metainfo_to_protobuf.hex()

    @classmethod
    def decode(self, hex: str) -> dict:
        '''decode card protobuf'''

        script = NulldataScript.unhexlify(hex).decompile().split(' ')[1]

        return parse_card_transfer_metainfo(bytes.fromhex(script),
                                            Settings.deck_version)

    @classmethod
    def simulate_issue(self, deckid: str=None, ncards: int=10,
                       verify: bool=False,
                       sign: str=False, send: bool=False) -> str:
        '''create a batch of simulated CardIssues on this deck'''

        receiver = [pa.Kutil(network=Settings.network).address for i in range(ncards)]
        amount = [random.randint(1, 100) for i in range(ncards)]

        return self.transfer(deckid=deckid, receiver=receiver, amount=amount,
                             verify=verify, sign=sign, send=send)

    def export(self, deckid: str, filename: str):
        '''export cards to csv'''

        cards = self.__list(deckid)['cards']
        export_to_csv(cards=list(cards), filename=filename)


class Transaction:

    def raw(self, txid: str) -> None:
        '''fetch raw tx and display it'''

        tx = provider.getrawtransaction(txid, 1)

        print(json.dumps(tx, indent=4))


def main():

    init_keystore()

    fire.Fire({
        'deck': Deck(),
        'card': Card(),
        'address': Address(),
        'transaction': Transaction(),
        'coin': Coin()
        })


if __name__ == '__main__':
    main()



'''default settings'''

default_conf = {
    "network": "tppc",
    "production": True,
    "deck_version": 1,  # deck version
    "change": "default",
    "provider": "explorer"  # explorer, cryptoid
    }



import csv
from pypeerassets.pautils import exponent_to_amount


def export_to_csv(cards, filename):
    '''export <cards> to csv <file>'''

    def format_card(card):
        '''filter out some info from CardTransfer'''

        c = card.__dict__.copy()
        c.pop("asset_specific_data")
        c["receiver"] = c["receiver"][0]
        c["amount"] = exponent_to_amount(c["amount"][0], card.number_of_decimals)
        c.pop("number_of_decimals")
        return c

    with open(filename, 'w') as csv_file:
        writer = csv.writer(csv_file, delimiter=';')
        writer.writerow(format_card(cards[0]).keys())  # writing the header
        for i in cards:
            writer.writerow(format_card(i).values())



from pypeerassets import RpcNode, Cryptoid, Explorer, pautils
from pacli.config import Settings


def set_up():
    '''setup'''

    # if provider is local node, check if PA P2TH is loaded in local node
    # this handles indexing of transaction
    if Settings.provider == "rpcnode":
        if Settings.production:
            if not provider.listtransactions("PAPROD"):
                pautils.load_p2th_privkeys_into_local_node(provider)
        if not Settings.production:
            if not provider.listtransactions("PATEST"):
                pautils.load_p2th_privkeys_into_local_node(provider, prod=False)


def configured_provider(Settings):
    " resolve settings into configured provider "

    if Settings.provider.lower() == "rpcnode":
        _provider = RpcNode

    elif Settings.provider.lower() == "cryptoid":
        _provider = Cryptoid

    elif Settings.provider.lower() == "explorer":
        _provider = Explorer

    else:
        raise Exception('invalid provider.')

    provider = _provider(network=Settings.network)
    set_up()

    return provider


provider = configured_provider(Settings)



from appdirs import user_config_dir
#import logging
import configparser
import os
from .keystore import load_key
from pypeerassets.pa_constants import param_query
from pacli.default_conf import default_conf
from pypeerassets import Kutil


def write_default_config(conf_file=None):
    '''if config file is not found, write a new one'''

    print("writing default config")
    config = configparser.ConfigParser()
    config["settings"] = default_conf
    if not conf_file:
        config.write()
    else:
        with open(conf_file, 'w') as configfile:
            config.write(configfile)


required = {"network", "deck_version", "production", "change", "provider"}


def read_conf(conf_file):

    config = configparser.ConfigParser()
    config.read(conf_file)
    try:
        settings = dict(config["settings"])
        assert set(settings.keys()).issuperset(required)

    except:
        print("config is outdated, saving current default config to", conf_file + ".sample")
        write_default_config(conf_file + ".sample")
        raise

    if settings["network"].startswith("t"):
        settings["testnet"] = True

    settings['p2th_address'] = param_query(settings['network']).P2TH_addr

    return settings


conf_dir = user_config_dir("pacli")
conf_file = os.path.join(conf_dir, "pacli.conf")
logfile = os.path.join(conf_dir, "pacli.log")


def init_config():
    '''if first run, setup local configuration directory.'''

    if not os.path.isdir(conf_dir):
        os.makedirs(conf_dir)
    if not os.path.exists(conf_file):
        write_default_config(conf_file)


def load_conf():
    '''load user configuration'''

    init_config()

    class Settings:
        pass

    settings = read_conf(conf_file)

    for key in settings:
        setattr(Settings, key, settings[key])

    setattr(Settings, 'deck_version', int(Settings.deck_version))
    setattr(Settings, 'key', Kutil(network=settings['network'],
                                   privkey=bytearray.fromhex(load_key())
                                   )
            )

    if settings['change'] == "default":
        Settings.change = Settings.key.address

    #logging.basicConfig(filename=logfile, level=logging.getLevelName(Settings.loglevel))
    #logging.basicConfig(level=logging.getLevelName(Settings.loglevel),
    #                    format="%(asctime)s %(levelname)s %(message)s")

    #logging.debug("logging initialized")

    return Settings


Settings = load_conf()



from setuptools import setup
from os import path

classifiers = [
  'Development Status :: 4 - Beta',
  'Intended Audience :: Financial and Insurance Industry',
  'Programming Language :: Python',
  'Operating System :: OS Independent',
  'Natural Language :: English',
  'License :: OSI Approved :: BSD License'
]

# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='pacli',
      version='0.4.5',
      description='Simple CLI PeerAssets client.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      keywords=['peerassets', 'blockchain', 'assets', 'client'],
      url='https://github.com/peerassets/pacli',
      author='Peerchemist',
      author_email='peerchemist@protonmail.ch',
      license='GPL',
      packages=['pacli'],
      install_requires=['pypeerassets', 'terminaltables',
                        'appdirs', 'fire', 'keyring',
                        ],
      entry_points={
          'console_scripts': [
              'pacli = pacli.__main__:main'
          ]}
      )



from decimal import Decimal
from typing import Union

from pypeerassets.exceptions import RecieverAmountMismatch
from pypeerassets.networks import net_query
from pypeerassets.transactions import (tx_output,
                                       p2pkh_script,
                                       nulldata_script,
                                       make_raw_transaction,
                                       Locktime)

from pacli.provider import provider
from pacli.config import Settings
from pacli.utils import sign_transaction, sendtx


class Coin:

    def sendto(self, address: Union[str], amount: Union[float],
               locktime: int=0) -> str:
        '''send coins to address'''

        if not len(address) == amount:
            raise RecieverAmountMismatch

        network_params = net_query(Settings.network)

        inputs = provider.select_inputs(Settings.key.address, sum(amount))

        outs = []

        for addr, index, amount in zip(address, range(len(address)), amount):
            outs.append(
                tx_output(network=Settings.network, value=Decimal(amount),
                          n=index,
                          script=p2pkh_script(address=addr,
                                              network=Settings.network))
            )

        #  first round of txn making is done by presuming minimal fee
        change_sum = Decimal(inputs['total'] - network_params.min_tx_fee)

        outs.append(
            tx_output(network=provider.network,
                      value=change_sum, n=len(outs)+1,
                      script=p2pkh_script(address=Settings.key.address,
                                          network=provider.network))
            )

        unsigned_tx = make_raw_transaction(network=provider.network,
                                           inputs=inputs['utxos'],
                                           outputs=outs,
                                           locktime=Locktime(locktime)
                                           )

        signedtx = sign_transaction(provider, unsigned_tx, Settings.key)

        return sendtx(signedtx)

    def opreturn(self, string: hex, locktime: int=0) -> str:
        '''send op_return transaction'''

        network_params = net_query(Settings.network)

        inputs = provider.select_inputs(Settings.key.address, 0.01)

        outs = [tx_output(network=provider.network,
                          value=Decimal(0), n=1,
                          script=nulldata_script(bytes.fromhex(string))
                          )
                ]

        #  first round of txn making is done by presuming minimal fee
        change_sum = Decimal(inputs['total'] - network_params.min_tx_fee)

        outs.append(
            tx_output(network=provider.network,
                      value=change_sum, n=len(outs)+1,
                      script=p2pkh_script(address=Settings.key.address,
                                          network=provider.network))
                    )

        unsigned_tx = make_raw_transaction(network=provider.network,
                                           inputs=inputs['utxos'],
                                           outputs=outs,
                                           locktime=Locktime(locktime)
                                           )

        signedtx = sign_transaction(provider, unsigned_tx, Settings.key)

        return sendtx(signedtx)



from pypeerassets.transactions import (MutableTransaction,
                                       sign_transaction
                                       )

from pacli.provider import provider
from pacli.config import Settings


def cointoolkit_verify(hex: str) -> str:
    '''tailor cointoolkit verify URL'''

    base_url = 'https://indiciumfund.github.io/cointoolkit/'
    if provider.network == "peercoin-testnet":
        mode = "mode=peercoin_testnet"
    if provider.network == "peercoin":
        mode = "mode=peercoin"

    return base_url + "?" + mode + "&" + "verify=" + hex


def signtx(rawtx: MutableTransaction) -> str:
    '''sign the transaction'''

    return sign_transaction(provider, rawtx, Settings.key)


def sendtx(signed_tx: MutableTransaction) -> str:
    '''send raw transaction'''

    provider.sendrawtransaction(signed_tx.hexlify())

    return signed_tx.txid



import keyring
from os import urandom
from btcpy.structs.crypto import PrivateKey


def generate_key() -> PrivateKey:
    '''generate new random key'''

    return urandom(32).hex()


def init_keystore() -> None:
    '''save key to the keystore'''

    if not keyring.get_password('pacli', 'key'):
        keyring.set_password("pacli", 'key', generate_key())


def load_key() -> PrivateKey:
    '''load key from the keystore'''

    init_keystore()

    key = keyring.get_password('pacli', 'key')

    return key
